﻿/*
aardio-electron 专用 websocket/json-rpc/aasdl 客户端
可以让aardio,js在electron主进程、渲染进程都可以相互调用函数
关于 aasdl : http://bbs.aardio.com/doc/aasdl/
*/
var startEnviron = JSON.parse(process.env.aarStartEnviron);

function isRendererProcess() {
  if (!process) return true;
  if (!process.type) return false;
  return process.type === "renderer";
}

let ws;
if (!isRendererProcess()) {
  require("module").globalPaths.push(startEnviron.nodePath);

  const { app } = require("electron");
  let WebSocket = require("ws.asar");
  ws = new WebSocket(startEnviron.rpcMainServerUrl);

  if (typeof startEnviron.args === "object") {
    for (var k in startEnviron.args) {
      app.commandLine.appendSwitch(k, startEnviron.args[k] + "");
    }
  }
} else {
  ws = new WebSocket(startEnviron.rpcServerUrl);
}

let events = require("events");
let eventEmitter = new events.EventEmitter();
let rpcNotifications = new Object();
let xcall;

ws.onopen = function(e) {

};
ws.onclose = function(e) {};
ws.onerror = function(e) {
  console.error(e);
};
ws.onmessage = function(e) {
  var rep = JSON.parse(e.data);
  if (typeof rep.method == "string") {
    var notify = rpcNotifications[rep.method];
    if (notify) {
      var result = notify.apply(ws, rep.params);

      if (rep.id) {
        var clientRep = JSON.stringify({
          id: rep.id,
          jsonrpc: "2.0",
          result: result
        });
        ws.send(clientRep);
      }
    }
  }
};

function on(method, notify) {
  eventEmitter.on(method, notify);
  rpcNotifications[method] = notify;
}

on("doScript",(js)=>{
  global.eval ("(()=>{" + js + "})()"); 
})

on("rpcClientId",(id)=>{
  exports.rpcClientId = id;
  xcall("?").then(function(aasdl, error) {
    var paseraasdl = function(obj, ex, pk) {
      for (var k in obj) {
        let method = k;
        if (typeof pk == "string") method = pk + "." + k;

        if (typeof obj[k] == "object") {
          ex[k] = {};
          paseraasdl(obj[k], ex[k], method);
          continue;
        }

        ex[k] = function() {
          return xcall.apply(ex, [method, ...arguments]);
        };
      }
    }

    paseraasdl(JSON.parse(aasdl), exports);

    if (isRendererProcess()) xcall("onLoadUrl", document.location.href);
    else {
      let win;
      const { app, BrowserWindow } = require("electron");
      createWindow = function() {
        let aardio = exports;
        win = new BrowserWindow(startEnviron.browserWindow);
        aardio.hwndElectron = win.getNativeWindowHandle().readInt32LE();
        aardio.main_initNativeWindow(aardio.hwndElectron);

        global.mainNativeWindow = win;
        if (startEnviron.extensions) {
          startEnviron.extensions.forEach(ext =>
            BrowserWindow.addExtension(ext)
          );
        }

        win.on("closed", () => {
          ws = null;
          global.mainNativeWindow = null;
          win = null;
        });

        win.once("ready-to-show", () => {
          win.show(); //如果在外部主动调用win.show()那么这个函数根本不会触发
        });

        eventEmitter.emit("ready", win);
        if (startEnviron.indexUrl)
          win.loadURL(startEnviron.indexUrl, startEnviron.indexOptions);
        aardio.main_readyNativeWindow(aardio.hwndElectron);
      };

      if (!global.mainNativeWindow) {
        if (app.isReady()) createWindow();
        else app.on("ready", createWindow);
      } else {
        eventEmitter.emit("ready", global.mainNativeWindow);
      }
    }
  }).catch(e =>{
    console.error(e)
  });
})

function off(method, notify) {
  eventEmitter.off(method, notify);
  delete rpcNotifications[method];
}

let rpchttpServerUrl = startEnviron.rpchttpServerUrl;
if (!isRendererProcess()) {
  var http = require("http");
  xcall = function (method, ...args) {
    return new Promise((resolve, reject) => {
      var reqData = JSON.stringify({
        method: method,
        id: method,
        jsonrpc: "2.0",
        params: args
      });
       
      var options = {
        host: "127.0.0.1",
        port: startEnviron.rpcPort,
        path: "/rpc/http",
        method: "POST",
        headers: {
          "Connection": "close",
          "Content-Type": "application/json",
          "Content-Length": Buffer.byteLength(reqData,'utf8'),
          "Rpc-Client-Id": exports.rpcClientId + ""
        }
      };

      var req = http.request(options, function(res) {
        var rawData = "";
        res.on("data", function(chunk) {
          rawData += chunk;
        });

        res.on("end", function() { 
          if(! (rawData && rawData.length) ){ 
            console.error("调用aardio函数时返回空数据,请求数据:",reqData)
            return reject( "RPC服务端返回空数据" )
          }
          var repData = JSON.parse(rawData)
          if (repData.error) return reject( repData.error );
          else return resolve(repData.result);
        });
      });
      req.shouldKeepAlive = false

      req.on('error', (e) => {
        console.error("调用aardio函数时遇到错误,请求数据:",reqData)
        reject( e );
      });

      req.write(reqData);
      req.end();
    });
  }
  
} else {
  xcall = async function(method, ...args) {
    var reqData = JSON.stringify({
      method: method,
      id: method,
      jsonrpc: "2.0",
      params: args
    });

    var repData = await fetch(rpchttpServerUrl, {
      body: reqData,
      cache: "no-cache",
      headers: {
        "Content-Type": "application/json",
        "Rpc-Client-Id": exports.rpcClientId + ""
      },
      method: "POST",
      mode: "cors"
    }).then(response => response.json())

    if (repData.error) {
      console.error("调用aardio函数时遇到错误,请求数据:",reqData);
      throw repData.error;
    }
    return repData.result;
  }
}

exports.call = xcall;
exports.on = on;
exports.startEnviron = startEnviron;
